function [output] = IA_PT_sampler(data, target, g, gprior, mode, RAPA, hparam, tau, nu, burnin, numbofits, thin, numbofreps, heat)

rhostar = zeros(1, length(heat) - 1);
for i = 1:(length(heat)-1)
    rhostar(i) = log(log(heat(i)/heat(i+1)));
end
order = zeros(numbofreps, length(heat));
for rep = 1:numbofreps
    order(rep, :) = 1:length(heat);
end

[n p] = size(data);

numbofchains = length(heat);

if ( length(hparam) == 1 )
    fixed = 1;
    w = hparam * ones(numbofreps, numbofchains);
    wstar = hparam;
else
    fixed = 0;
    wa = hparam(1);
    wb = hparam(2);
    wstar = wa / (wa + wb);
end

logita = 0.1 / p;
logitb = 1 - 0.1 / p;

loglike = zeros(numbofreps, numbofchains);

gamma = zeros(numbofreps, p, numbofchains);
for chain = 1:numbofchains
    for rep = 1:numbofreps
        gamma(rep, :, chain) = rand(1, p) < wstar;
        
        datastar = [ones(n, 1) data(:, gamma(rep, :, chain)==1)];
        if ( gprior == 1 )
            n0star = zeros(sum(gamma(rep, :, chain)));
            n0star(2:sum(gamma(rep, :, chain)), 2:sum(gamma(rep, :, chain))) = 1 / g  * datastar(:, 2:end)' * datastar(:, 2:end);
        else
            n0star = 1 / g * eye(sum(gamma(rep, :, chain)) + 1);
            n0star(1, 1) = 0;
        end
        
        loglike(rep, chain) = - 0.5 * sum(gamma(rep, :, chain)) * log(g) - 0.5 * log(det(datastar' * datastar + n0star));
        loglike(rep, chain) = loglike(rep, chain) - 0.5 * n * log(target' * target - target' * datastar * inv(datastar' * datastar + n0star) * datastar' * target);
    end
end

zetaAS = nu / ((1 - wstar) * p) * ones(numbofchains, p);
zetaDS = nu / (wstar * p) * ones(numbofchains, p);
logitzetaAS = log(zetaAS - logita) - log(logitb - zetaAS);
logitzetaDS = log(zetaDS - logita) - log(logitb - zetaDS);

totalaccept = zeros(1, numbofchains);
totalcount = zeros(1, numbofchains);

totalaccept2 = zeros(1, numbofchains - 1);
totalcount2 = zeros(1, numbofchains - 1);

sumgamma = zeros(1, p);
holdmodelsize = zeros(numbofits, numbofreps);
holdloglike = zeros(numbofits, numbofreps);
holdaccept = zeros(numbofits, 1);
if ( mode == 2 )
    holdgamma = zeros(numbofreps, numbofits, p);
else
    holdgamma = [];
end

for it = 1:(burnin+numbofits*thin)
       
    if ( mod(it, 100) == 0 )
        disp(['it = ' num2str(it)]);
        disp(['numbofreps = ' num2str(numbofreps)]);
        disp(['accept = ' num2str(totalaccept./totalcount)]);
        disp(num2str(heat));
        disp(num2str(totalaccept2./totalcount2));
        disp(' ');
    end
    
    for chain = 1:numbofchains
        for rep = 1:numbofreps
            chain1 = order(rep, chain);
            
            check1 = logitzetaAS(chain, :)<0;
            check2 = logitzetaDS(chain, :)<0;
            zetaAS(check1==1) = (logita + logitb * exp(logitzetaAS(chain1, check1==1))) ./ (1 + exp(logitzetaAS(chain1, check1==1)));
            zetaDS(check2==1) = (logita + logitb * exp(logitzetaDS(chain1, check2==1))) ./ (1 + exp(logitzetaDS(chain1, check2==1)));
            zetaAS(check1==0) = (logita * exp(- logitzetaAS(chain1, check1==0)) + logitb) ./ (exp(- logitzetaAS(chain1, check1==0)) + 1);
            zetaDS(check2==0) = (logita * exp(- logitzetaDS(chain1, check2==0)) + logitb) ./ (exp(- logitzetaDS(chain1, check2==0)) + 1);
            
            gammaA = zeros(1, p);
            gammaD = zeros(1, p);
            newgamma = gamma(rep, :, chain);
            logprob = 0;
            
            gammaA(gamma(rep, :, chain)==0) = rand(1, sum(gamma(rep, :, chain)==0)) < zetaAS(gamma(rep, :, chain)==0);
            newgamma(gammaA==1) = 1;
            gammaD(gamma(rep, :, chain)==1) = rand(1, sum(gamma(rep, :, chain)==1)) < zetaDS(gamma(rep, :, chain)==1);
            newgamma(gammaD==1) = 0;
            temp = (gamma(rep, :, chain)==0) .* (newgamma==1);
            logprob = logprob - sum(log(zetaAS(temp==1)));
            temp = (gamma(rep, :, chain)==0) .* (newgamma==0);
            logprob = logprob - sum(log(1 - zetaAS(temp==1)));
            temp = (gamma(rep, :, chain)==1) .* (newgamma==0);
            logprob = logprob - sum(log(zetaDS(temp==1)));
            temp = (gamma(rep, :, chain)==1) .* (newgamma==1);
            logprob = logprob - sum(log(1 - zetaDS(temp==1)));
            temp = (newgamma==0) .* (gamma(rep, :, chain)==1);
            logprob = logprob + sum(log(zetaAS(temp==1)));
            temp = (newgamma==0) .* (gamma(rep, :, chain)==0);
            logprob = logprob + sum(log(1 - zetaAS(temp==1)));
            temp = (newgamma==1) .* (gamma(rep, :, chain)==0);
            logprob = logprob + sum(log(zetaDS(temp==1)));
            temp = (newgamma==1) .* (gamma(rep, :, chain)==1);
            logprob = logprob + sum(log(1 - zetaDS(temp==1)));
            change = sum(gammaA) + sum(gammaD);
            
            datastar = [ones(n, 1) data(:, newgamma==1)];
            if ( gprior == 1 )
                n0star = zeros(sum(newgamma));
                n0star(2:sum(newgamma), 2:sum(newgamma)) = 1 / g  * datastar(:, 2:end)' * datastar(:, 2:end);
            else
                n0star = 1 / g * eye(sum(newgamma) + 1);
                n0star(1, 1) = 0;
            end
            
            newloglike = - 0.5 * sum(newgamma) * log(g) - 0.5 * log(det(datastar' * datastar + n0star));
            newloglike = newloglike - 0.5 * n * log(target' * target - target' * datastar * inv(datastar' * datastar + n0star) * datastar' * target);
            
            logaccept = heat(chain1) * (newloglike - loglike(rep, chain));
            if ( fixed == 1 )
                logaccept = logaccept + sum(newgamma) * log(w(rep, chain)) + (p - sum(newgamma)) * log(1 - w(rep, chain));
                logaccept = logaccept - sum(gamma(rep, :, chain)) * log(w(rep, chain)) - (p - sum(gamma(rep, :, chain))) * log(1 - w(rep, chain));
            else
                logaccept = logaccept + gammaln(wa + sum(newgamma)) + gammaln(wb + p - sum(newgamma));
                logaccept = logaccept - gammaln(wa + sum(gamma(rep, :, chain))) - gammaln(wb + p - sum(gamma(rep, :, chain)));
            end
            logaccept = logaccept + logprob;
            
            accept = 1;
            if ( (isnan(logaccept) == 1) || (isinf(logaccept) == 1) )
                accept = 0;
            elseif ( logaccept < 0 )
                accept = exp(logaccept);
            end
            
            if ( rand < accept )
                gamma(rep, :, chain) = newgamma;
                loglike(rep, chain) = newloglike;
            end
            
            totalaccept(chain1) = totalaccept(chain1) + (change~=0)*accept;
            totalcount(chain1) = totalcount(chain1) + 1;

            logaccept2 = - logaccept;
            accept2 = 1;
            if ( (isnan(logaccept2) == 1) || (isinf(logaccept2) == 1) )
                accept2 = 0;
            elseif ( logaccept2 < 0 )
                accept2 = exp(logaccept2);
            end
            
            logitzetaAS(chain1, :) = logitzetaAS(chain1, :) + 2 / it^0.55 * (change~=0) * (accept - tau) .* gammaA;
            if ( RAPA == 1 )
                logitzetaAS(chain1, :) = logitzetaAS(chain1, :) + 2 / it^0.55 * (change~=0) * (accept - tau) .* gammaA * (1 - accept);
                logitzetaDS(chain1, :) = logitzetaDS(chain1, :) + 2 / it^0.55 * (change~=0) * (accept2 - tau) .* gammaA * accept;
            end
            
            logitzetaDS(chain1, :) = logitzetaDS(chain1, :) + 2 / it^0.55 * (change~=0) * (accept - tau) .* gammaD;
            if ( RAPA == 1 )
                logitzetaDS(chain1, :) = logitzetaDS(chain1, :) + 2 / it^0.55 * (change~=0) * (accept - tau) .* gammaD * (1 - accept);
                logitzetaAS(chain1, :) = logitzetaAS(chain1, :) + 2 / it^0.55 * (change~=0) * (accept2 - tau) .* gammaD * accept;
            end
        end
    end
    
    if ( length(heat) > 1 )
        % Swap move
        for rep = 1:numbofreps
            check = 1;
            while ( check == 1 )
                chain1 = ceil(rand * length(heat));
                check = (order(rep, chain1) == length(heat));
            end
            chain2 = find(order(rep, :) == order(rep, chain1) + 1);
            
            loglike1 = loglike(rep, chain1);
            loglike2 = loglike(rep, chain2);
            
            logaccept = heat(order(rep, chain1)) * loglike2 + heat(order(rep, chain2)) * loglike1 - heat(order(rep, chain1)) * loglike1 - heat(order(rep, chain2)) * loglike2;
            
            accept = 1;
            if ( logaccept  < 0 )
                accept = exp(logaccept);
            end
            totalaccept2(order(rep, chain1)) = totalaccept2(order(rep, chain1)) + accept;
            totalcount2(order(rep, chain1)) = totalcount2(order(rep, chain1)) + 1;
            
            rhostar(order(rep, chain1)) = rhostar(order(rep, chain1)) + 1 / it^0.55 * (accept - 0.234);
            if ( rhostar(order(rep, chain1)) > 4 )
                rhostar(order(rep, chain1)) = 4;
            end
            
            for j = 1:(length(heat)-1)
                heat(j+1) = heat(j) * exp(- exp(rhostar(j)));
            end
            
            if ( rand < accept )
                order(rep, chain1) = order(rep, chain1) + 1;
                order(rep, chain2) = order(rep, chain2) - 1;
            end
        end
    end
    
    if ( (it > burnin) && (mod(it - burnin, thin) == 0) )
 
        for rep = 1:numbofreps
            top_chain = find(order(rep, :) == 1);
            sumgamma = sumgamma + gamma(rep, :, top_chain);
            if ( fixed == 1 )
                holdloglike((it - burnin) / thin, rep) = loglike(rep, top_chain) + sum(gamma(rep, :, top_chain)) * log(w(rep, chain)) + sum(1-gamma(rep, :, top_chain)) * log(1 - w(rep, chain));
            else
                holdloglike((it - burnin) / thin, rep) = loglike(rep, top_chain) + gammaln(wa + sum(gamma(rep, :, top_chain))) + gammaln(wb + p - sum(gamma(rep, :, top_chain))) - gammaln(wa + wb + p);
            end               
            holdmodelsize((it - burnin) / thin, rep) = sum(gamma(rep, :, top_chain));
            holdaccept((it - burnin) / thin) = totalaccept(top_chain) / totalcount(top_chain);
            if ( mode == 2 )
                holdgamma(rep, (it - burnin) / thin, :) = gamma(rep, :, 1);
            end
        end
    end
end

max(sumgamma)

sumgamma = sumgamma / (numbofits * numbofreps);
zetaAS = (logita + logitb * exp(logitzetaAS)) ./ (1 + exp(logitzetaAS));
zetaDS = (logita + logitb * exp(logitzetaDS)) ./ (1 + exp(logitzetaDS));
accept2 = totalaccept2./totalcount2;

output = struct('prob_inclusion', sumgamma, 'logpost', holdloglike, 'modelsize', holdmodelsize, 'accept', holdaccept, 'gamma', holdgamma, 'zetaAS', zetaAS, 'zetaDS', zetaDS, 'accept2', accept2, 'heat', heat);