function [output] = IA_SMC_sampler(data, target, g, gprior, RAPA, hparam, tau, c_ESS, nu, numbofparts, numbofits)

gammat = 0;

maxgammat =  1-10^(-2);

[n p] = size(data);

w = 5 / p;


if ( length(hparam) == 1 )
    fixed = 1;
    w = hparam;
    wstar = hparam;
else
    fixed = 0;
    wa = hparam(1);
    wb = hparam(2);
    wstar = wa / (wa + wb);
end

logita = 0.1 / p;
logitb = 1 - 0.1 / p;

gamma = rand(p, numbofparts) < wstar;

loglike = zeros(1, numbofparts);
logprior = zeros(1, numbofparts);
for part = 1:numbofparts
    datastar = [ones(n, 1) data(:, gamma(:, part)==1)];
    if ( gprior == 1 )
        n0star = zeros(sum(gamma(:, part)));
        n0star(2:sum(gamma(:, part)), 2:sum(gamma(:, part))) = 1 / g  * datastar(:, 2:end)' * datastar(:, 2:end);
    else
        n0star = 1 / g * eye(sum(gamma(:, part)) + 1);
        n0star(1, 1) = 0;
    end
    
    loglike(part) = - 0.5 * sum(gamma(:, part)) * log(g) - 0.5 * log(det(datastar' * datastar + n0star));
    loglike(part) = loglike(part) - 0.5 * n * log(target' * target - target' * datastar * inv(datastar' * datastar + n0star) * datastar' * target);

    if ( fixed == 1 )
        logprior(part) = sum(gamma(:, part)) * log(w) + (p - sum(gamma(:, part))) * log(1 - w);
    else
        logprior(part) = gammaln(wa + sum(gamma(:, part))) + gammaln(wb + p - sum(gamma(:, part)));
    end
end

zetaAS = nu / ((1 - wstar) * p) * ones(1, p);
zetaDS = nu / (wstar * p) * ones(1, p);
logitzetaAS = log(zetaAS - logita) - log(logitb - zetaAS);
logitzetaDS = log(zetaDS - logita) - log(logitb - zetaDS);

partstar = zeros(1, numbofparts);

totalaccept = 0;
totalcount = 0;

target_ESS = c_ESS * numbofparts;

disp(['gammat = ' num2str(gammat)]);
while ( gammat < 1-10^(-2) )
    

    logweight = loglike * (maxgammat - gammat);
    
    weight = exp(logweight - max(logweight));
    weight = weight / sum(weight);
    ESS = sum(weight)^2 / sum(weight.^2);
    adj_ESS = ESS - target_ESS;
    
    if ( adj_ESS > 0 )
        newgammat = maxgammat;
    else
        x = fzero('Calc_new_ESS', 0, [], gammat, loglike, target_ESS);
        z = exp(x) / (1 + exp(x));
        newgammat = gammat + (1 - gammat) * z;
    end
%     newgammat = min(1, gammat + 0.1);
%     check = 0;
%     while ( check == 0 )
%         newgammat = newgammat - 0.5 * (newgammat - gammat);
%         logweight = loglike * (newgammat - gammat);
%         
%         weight = exp(logweight - max(logweight));
%         weight = weight / sum(weight);
%         ESS = sum(weight)^2 / sum(weight.^2);
%                 
%         check = (ESS > c_ESS * numbofparts);
%     end
    
    logweight = loglike * (newgammat - gammat);
    weight = exp(logweight - max(logweight));
    weight = weight / sum(weight);
    ESS = sum(weight)^2 / sum(weight.^2);
    
    fprob = cumsum(weight);
    fprob = fprob / fprob(end);
    
    u1 = rand / numbofparts;
    k = 1;
    it = 0;
    while ( u1 < 1 )
        while ( (fprob(k) < u1) && (k <= numbofparts) )
            k = k + 1;
        end
        it = it + 1;
        u1 = u1 + 1 / numbofparts;
        partstar(it) = k;
    end
    
    gamma = gamma(:, partstar);
    loglike = loglike(partstar);
    logprior = logprior(partstar);
    
    gammat = newgammat;
    
    disp(['gammat = ' num2str(gammat) ' ESS = ' num2str(ESS)]);
    
    for it = 1:numbofits
        disp(['it = ' num2str(it)]);
        for part = 1:numbofparts
            check1 = logitzetaAS<0;
            check2 = logitzetaDS<0;
            zetaAS(check1==1) = (logita + logitb * exp(logitzetaAS(check1==1))) ./ (1 + exp(logitzetaAS(check1==1)));
            zetaDS(check2==1) = (logita + logitb * exp(logitzetaDS(check2==1))) ./ (1 + exp(logitzetaDS(check2==1)));
            zetaAS(check1==0) = (logita * exp(- logitzetaAS(check1==0)) + logitb) ./ (exp(- logitzetaAS(check1==0)) + 1);
            zetaDS(check2==0) = (logita * exp(- logitzetaDS(check2==0)) + logitb) ./ (exp(- logitzetaDS(check2==0)) + 1);
            
            gammaA = zeros(1, p);
            gammaD = zeros(1, p);
            newgamma = gamma(:, part);
            logprob = 0;
            
            gammaA(gamma(:, part)==0) = rand(1, sum(gamma(:, part)==0)) < zetaAS(gamma(:, part)==0);
            newgamma(gammaA==1) = 1;
            gammaD(gamma(:, part)==1) = rand(1, sum(gamma(:, part)==1)) < zetaDS(gamma(:, part)==1);
            newgamma(gammaD==1) = 0;
            temp = (gamma(:, part)==0) .* (newgamma==1);
            logprob = logprob - sum(log(zetaAS(temp==1)));
            temp = (gamma(:, part)==0) .* (newgamma==0);
            logprob = logprob - sum(log(1 - zetaAS(temp==1)));
            temp = (gamma(:, part)==1) .* (newgamma==0);
            logprob = logprob - sum(log(zetaDS(temp==1)));
            temp = (gamma(:, part)==1) .* (newgamma==1);
            logprob = logprob - sum(log(1 - zetaDS(temp==1)));
            temp = (newgamma==0) .* (gamma(:, part)==1);
            logprob = logprob + sum(log(zetaAS(temp==1)));
            temp = (newgamma==0) .* (gamma(:, part)==0);
            logprob = logprob + sum(log(1 - zetaAS(temp==1)));
            temp = (newgamma==1) .* (gamma(:, part)==0);
            logprob = logprob + sum(log(zetaDS(temp==1)));
            temp = (newgamma==1) .* (gamma(:, part)==1);
            logprob = logprob + sum(log(1 - zetaDS(temp==1)));
            change = sum(gammaA) + sum(gammaD);
            
            datastar = [ones(n, 1) data(:, newgamma==1)];
            if ( gprior == 1 )
                n0star = zeros(sum(newgamma));
                n0star(2:sum(newgamma), 2:sum(newgamma)) = 1 / g  * datastar(:, 2:end)' * datastar(:, 2:end);
            else
                n0star = 1 / g * eye(sum(newgamma) + 1);
                n0star(1, 1) = 0;
            end

            newloglike = - 0.5 * sum(newgamma) * log(g) - 0.5 * log(det(datastar' * datastar + n0star));
            newloglike = newloglike - 0.5 * n * log(target' * target - target' * datastar * inv(datastar' * datastar + n0star) * datastar' * target);
            if ( fixed == 1 )
                newlogprior = sum(newgamma) * log(w) + (p - sum(newgamma)) * log(1 - w);
            else
                newlogprior = gammaln(wa + sum(newgamma)) + gammaln(wb + p - sum(newgamma));
            end        
            
            logaccept = gammat * (newloglike - loglike(part)) + newlogprior - logprior(part) + logprob;
            
            accept = 1;
            if ( (isnan(logaccept) == 1) || (isinf(logaccept) == 1) )
                accept = 0;
            elseif ( logaccept < 0 )
                accept = exp(logaccept);
            end
            
            if ( rand < accept )
                gamma(:, part) = newgamma;
                loglike(part) = newloglike;
                logprior(part) = newlogprior;
            end
            
            totalaccept = totalaccept + (change~=0)*accept;
            totalcount = totalcount + 1;
            
            logaccept2 = - logaccept;
            accept2 = 1;
            if ( (isnan(logaccept2) == 1) || (isinf(logaccept2) == 1) )
                accept2 = 0;
            elseif ( logaccept2 < 0 )
                accept2 = exp(logaccept2);
            end
            
            logitzetaAS = logitzetaAS + 2 / it^0.55 * (change~=0) * (accept - tau) .* gammaA;
            if ( RAPA == 1 )
                logitzetaAS = logitzetaAS + 2 / it^0.55 * (change~=0) * (accept - tau) .* gammaA * (1 - accept);
                logitzetaDS = logitzetaDS + 2 / it^0.55 * (change~=0) * (accept2 - tau) .* gammaA * accept;
            end
            
            logitzetaDS = logitzetaDS + 2 / it^0.55 * (change~=0) * (accept - tau) .* gammaD;
            if ( RAPA == 1 )
                logitzetaDS = logitzetaDS + 2 / it^0.55 * (change~=0) * (accept - tau) .* gammaD * (1 - accept);
                logitzetaAS = logitzetaAS + 2 / it^0.55 * (change~=0) * (accept2 - tau) .* gammaD * accept;
            end
        end
    end
    disp(['accept = ' num2str(totalaccept/totalcount)]);
end

gammat = 1;
disp(['gammat = ' num2str(gammat)]);
for it = 1:numbofits
    disp(['it = ' num2str(it)]);
    for part = 1:numbofparts
        check1 = logitzetaAS<0;
        check2 = logitzetaDS<0;
        zetaAS(check1==1) = (logita + logitb * exp(logitzetaAS(check1==1))) ./ (1 + exp(logitzetaAS(check1==1)));
        zetaDS(check2==1) = (logita + logitb * exp(logitzetaDS(check2==1))) ./ (1 + exp(logitzetaDS(check2==1)));
        zetaAS(check1==0) = (logita * exp(- logitzetaAS(check1==0)) + logitb) ./ (exp(- logitzetaAS(check1==0)) + 1);
        zetaDS(check2==0) = (logita * exp(- logitzetaDS(check2==0)) + logitb) ./ (exp(- logitzetaDS(check2==0)) + 1);

        gammaA = zeros(1, p);
        gammaD = zeros(1, p);
        newgamma = gamma(:, part);
        logprob = 0;
        
        gammaA(gamma(:, part)==0) = rand(1, sum(gamma(:, part)==0)) < zetaAS(gamma(:, part)==0);
        newgamma(gammaA==1) = 1;
        gammaD(gamma(:, part)==1) = rand(1, sum(gamma(:, part)==1)) < zetaDS(gamma(:, part)==1);
        newgamma(gammaD==1) = 0;
        temp = (gamma(:, part)==0) .* (newgamma==1);
        logprob = logprob - sum(log(zetaAS(temp==1)));
        temp = (gamma(:, part)==0) .* (newgamma==0);
        logprob = logprob - sum(log(1 - zetaAS(temp==1)));
        temp = (gamma(:, part)==1) .* (newgamma==0);
        logprob = logprob - sum(log(zetaDS(temp==1)));
        temp = (gamma(:, part)==1) .* (newgamma==1);
        logprob = logprob - sum(log(1 - zetaDS(temp==1)));
        temp = (newgamma==0) .* (gamma(:, part)==1);
        logprob = logprob + sum(log(zetaAS(temp==1)));
        temp = (newgamma==0) .* (gamma(:, part)==0);
        logprob = logprob + sum(log(1 - zetaAS(temp==1)));
        temp = (newgamma==1) .* (gamma(:, part)==0);
        logprob = logprob + sum(log(zetaDS(temp==1)));
        temp = (newgamma==1) .* (gamma(:, part)==1);
        logprob = logprob + sum(log(1 - zetaDS(temp==1)));
        change = sum(gammaA) + sum(gammaD);
        
        datastar = [ones(n, 1) data(:, newgamma==1)];
        if ( gprior == 1 )
            n0star = zeros(sum(newgamma));
            n0star(2:sum(newgamma), 2:sum(newgamma)) = 1 / g  * datastar(:, 2:end)' * datastar(:, 2:end);
        else
            n0star = 1 / g * eye(sum(newgamma) + 1);
            n0star(1, 1) = 0;
        end
        
        newloglike = - 0.5 * sum(newgamma) * log(g) - 0.5 * log(det(datastar' * datastar + n0star));
        newloglike = newloglike - 0.5 * n * log(target' * target - target' * datastar * inv(datastar' * datastar + n0star) * datastar' * target);
        if ( fixed == 1 )
            newlogprior = sum(newgamma) * log(w) + (p - sum(newgamma)) * log(1 - w);
        else
            newlogprior = gammaln(wa + sum(newgamma)) + gammaln(wb + p - sum(newgamma));
        end
        
        logaccept = gammat * (newloglike - loglike(part)) + newlogprior - logprior(part) + logprob;
        
        accept = 1;
        if ( (isnan(logaccept) == 1) || (isinf(logaccept) == 1) )
            accept = 0;
        elseif ( logaccept < 0 )
            accept = exp(logaccept);
        end
        
        if ( rand < accept )
            gamma(:, part) = newgamma;
            loglike(part) = newloglike;
            logprior(part) = newlogprior;
        end
        
        totalaccept = totalaccept + (change~=0)*accept;
        totalcount = totalcount + 1;
        
        logaccept2 = - logaccept;
        accept2 = 1;
        if ( (isnan(logaccept2) == 1) || (isinf(logaccept2) == 1) )
            accept2 = 0;
        elseif ( logaccept2 < 0 )
            accept2 = exp(logaccept2);
        end
        
        logitzetaAS = logitzetaAS + 2 / it^0.55 * (change~=0) * (accept - tau) .* gammaA;
        if ( RAPA == 1 )
            logitzetaAS = logitzetaAS + 2 / it^0.55 * (change~=0) * (accept - tau) .* gammaA * (1 - accept);
            logitzetaDS = logitzetaDS + 2 / it^0.55 * (change~=0) * (accept2 - tau) .* gammaA * accept;
        end
        
        logitzetaDS = logitzetaDS + 2 / it^0.55 * (change~=0) * (accept - tau) .* gammaD;
        if ( RAPA == 1 )
            logitzetaDS = logitzetaDS + 2 / it^0.55 * (change~=0) * (accept - tau) .* gammaD * (1 - accept);
            logitzetaAS = logitzetaAS + 2 / it^0.55 * (change~=0) * (accept2 - tau) .* gammaD * accept;
        end
    end
end
disp(['accept = ' num2str(totalaccept/totalcount)]);

sumgamma = mean(gamma, 2);
modelsize = sum(gamma);

output = struct('prob_inclusion', sumgamma, 'logpost', loglike, 'modelsize', modelsize, 'gamma', gamma, 'zetaAS', zetaAS, 'zetaDS', zetaDS);
